#![allow(non_snake_case)]

use std::ops::{Add, Sub, Mul, Div, BitOr};
use std::f64::consts::*;

/// 2D Vector Representation
#[derive(Debug, Copy, Clone)]
pub struct Vector2D {
    /// `X` axis
    pub X: f64,
    /// `Y` axis
    pub Y: f64,
}

pub trait VectorOperations {
    fn magnitude(&self) -> f64;
    fn normalize(&self) -> Vector2D;
    fn angle(lhs: Vector2D, rhs: Vector2D) -> f64;
    fn dot_product(lhs: Vector2D, rhs: Vector2D) -> f64;
}

impl Vector2D {
    /// Returns a `Vector2D` with its `X` and `Y` axis initialize.
    ///
    /// # Arguments
    ///
    /// * `x` - Initializing value for X axis.
    /// * `y` - Initializing value for Y axis.
    ///
    /// # Examples
    /// ```
    /// use MBrane::ts::Vector2D::*;
    /// let vector = Vector2D::new(8.5, 12.3);
    /// ```
    pub fn new(x: f64, y: f64) -> Vector2D {
        Vector2D { X: x, Y: y }
    }

    /// Perform Vector Normalization on `self`.
    /// # Arguments
    ///
    /// * `&mut self` - A reference to mutable `self`.
    ///
    /// # Examples
    /// ```
    /// use MBrane::ts::Vector2D::*;
    /// let mut vector = Vector2D::new(16.5, 3.14);
    /// vector.self_normalize();
    /// println!("{:?}", vector);
    /// ```
    pub fn self_normalize(&mut self) {
        let magnitude = self.magnitude();
        self.X = self.X / magnitude;
        self.Y = self.Y / magnitude;
    }
}

impl VectorOperations for Vector2D {
    /// Returns the Vector magnitude/length of `self`.
    /// # Arguments
    ///
    /// * `&self` - A reference to `self`.
    ///
    /// # Examples
    /// ```
    /// use MBrane::ts::Vector2D::*;
    /// let vector = Vector2D::new(415.0, -20.0);
    /// println!("{:?}", vector.magnitude());
    /// ```
    fn magnitude(&self) -> f64 {
        (((self.X * self.X) + (self.Y * self.Y))).sqrt()
    }

    /// Rerturns a normalized `Vector2D`.
    /// # Arguments
    ///
    /// * `&self` - A reference to `self`.
    ///
    /// # Examples
    /// ```
    /// use MBrane::ts::Vector2D::*;
    /// let vector = Vector2D::new(16.5, 3.14);
    /// let normalized_vector = vector.normalize();
    /// println!("{:?}", normalized_vector);
    /// ```
    fn normalize(&self) -> Vector2D {
        let magnitude = self.magnitude();
        Vector2D::new(self.X / magnitude, self.Y / magnitude)
    }

    /// Return the angle between two `Vector2D`.
    /// # Arguments
    ///
    /// * `lhs` - Left hand side vector.
    /// * `rhs` - Right hand side vector.
    ///
    /// # Examples
    /// ```
    /// use MBrane::ts::Vector2D::*;
    /// let lhs = Vector2D::new(6.0, 8.0);
    /// let rhs = Vector2D::new(5.0, 12.0);
    /// let angle = Vector2D::angle(lhs, rhs);
    /// println!("{:?}", angle);
    /// ```
    fn angle(lhs: Vector2D, rhs: Vector2D) -> f64 {
        ((lhs | rhs) / (lhs.magnitude() * rhs.magnitude())).acos() * (180.0 / PI)
    }

    /// Return the the dot product of two `Vector2D`
    /// # Arguments
    ///
    /// * `lhs` - Left hand side vector.
    /// * `rhs` - Right hand side vector.
    ///
    /// # Examples
    /// ```
    /// use MBrane::ts::Vector2D::*;
    /// let lhs = Vector2D::new(1.85, 35.6);
    /// let rhs = Vector2D::new(8.5, 42.0);
    /// println!("{:?}", Vector2D::dot_product(lhs, rhs));
    /// ```
    fn dot_product(lhs: Vector2D, rhs: Vector2D) -> f64 {
        (lhs.magnitude() * rhs.magnitude()) * (Vector2D::angle(lhs, rhs)).cos()
    }
}

impl Add for Vector2D {
    type Output = Vector2D;

    fn add(self, rhs: Vector2D) -> Vector2D {
        Vector2D::new(self.X + rhs.X, self.Y + rhs.Y)
    }
}

impl Sub for Vector2D {
    type Output = Vector2D;

    fn sub(self, rhs: Vector2D) -> Vector2D {
        Vector2D::new(self.X - rhs.X, self.Y - rhs.Y)
    }
}

impl Mul for Vector2D {
    type Output = Vector2D;

    fn mul(self, rhs: Vector2D) -> Vector2D {
        Vector2D::new(self.X * rhs.X, self.Y * rhs.Y)
    }
}

impl Div for Vector2D {
    type Output = Vector2D;

    fn div(self, rhs: Vector2D) -> Vector2D {
        Vector2D::new(self.X / rhs.X, self.Y / rhs.Y)
    }
}

impl BitOr for Vector2D {
    type Output = f64;

    fn bitor(self, rhs: Vector2D) -> f64 {
        (self.X * rhs.X) + (self.Y * rhs.Y)
    }
}
