#![allow(non_snake_case)]

use std::ops::{Add, Sub, Mul, Div, BitOr};
use std::f64::consts::*;

/// 3D Vector Representation
#[derive(Debug, Copy, Clone)]
pub struct Vector3D {
    /// `X` axis
    pub X: f64,
    /// `Y` axis
    pub Y: f64,
    /// `Z` axis
    pub Z: f64,
}

pub trait VectorOperations {
    fn magnitude(&self) -> f64;
    fn normalize(&self) -> Vector3D;
    fn angle(lhs: Vector3D, rhs: Vector3D) -> f64;
    fn dot_product(lhs: Vector3D, rhs: Vector3D) -> f64;
}

impl Vector3D {
    /// Returns a `Vector3D` with its `X`, `Y` and `Z` axis initialize.
    ///
    /// # Arguments
    ///
    /// * `x` - Initializing value for X axis.
    /// * `y` - Initializing value for Y axis.
    /// * `z` - Initializing value for Z axis.
    ///
    /// # Examples
    /// ```
    /// use MBrane::ts::Vector3D::*;
    /// let vector = Vector3D::new(8.5, 12.3, 18.0);
    /// ```
    pub fn new(x: f64, y: f64, z: f64) -> Vector3D {
        Vector3D { X: x, Y: y, Z: z }
    }

    /// Perform Vector Normalization on `self`.
    /// # Arguments
    ///
    /// * `&mut self` - A reference to mutable `self`.
    ///
    /// # Examples
    /// ```
    /// use MBrane::ts::Vector3D::*;
    /// let mut vector = Vector3D::new(16.5, 3.14, 2.5);
    /// vector.self_normalize();
    /// println!("{:?}", vector);
    /// ```
    pub fn self_normalize(&mut self) {
        let magnitude = self.magnitude();
        self.X = self.X / magnitude;
        self.Y = self.Y / magnitude;
        self.Z = self.Z / magnitude;
    }
}

impl VectorOperations for Vector3D {
    /// Returns the Vector magnitude/length of `self`.
    /// # Arguments
    ///
    /// * `&self` - A reference to `self`.
    ///
    /// # Examples
    /// ```
    /// use MBrane::ts::Vector3D::*;
    /// let vector = Vector3D::new(415.0, -20.0, 8.12);
    /// println!("{:?}", vector.magnitude());
    /// ```
    fn magnitude(&self) -> f64 {
        (((self.X * self.X) + (self.Y * self.Y) + (self.Z * self.Z))).sqrt()
    }

    /// Rerturns a normalized `Vector3D`.
    /// # Arguments
    ///
    /// * `&self` - A reference to `self`.
    ///
    /// # Examples
    /// ```
    /// use MBrane::ts::Vector3D::*;
    /// let vector = Vector3D::new(16.5, 3.14, 18.0);
    /// let normalized_vector = vector.normalize();
    /// println!("{:?}", normalized_vector);
    /// ```
    fn normalize(&self) -> Vector3D {
        let magnitude = self.magnitude();
        Vector3D::new(self.X / magnitude, self.Y / magnitude, self.Z / magnitude)
    }

    /// Return the angle between two `Vector3D`.
    /// # Arguments
    ///
    /// * `lhs` - Left hand side vector.
    /// * `rhs` - Right hand side vector.
    ///
    /// # Examples
    /// ```
    /// use MBrane::ts::Vector3D::*;
    /// let lhs = Vector3D::new(6.0, 8.0, 25.0);
    /// let rhs = Vector3D::new(5.0, 12.0, 31.3);
    /// let angle = Vector3D::angle(lhs, rhs);
    /// println!("{:?}", angle);
    /// ```
    fn angle(lhs: Vector3D, rhs: Vector3D) -> f64 {
        ((lhs | rhs) / (lhs.magnitude() * rhs.magnitude())).acos() * (180.0 / PI)
    }

    /// Return the the dot product of two `Vector3D`
    /// # Arguments
    ///
    /// * `lhs` - Left hand side vector.
    /// * `rhs` - Right hand side vector.
    ///
    /// # Examples
    /// ```
    /// use MBrane::ts::Vector3D::*;
    /// let lhs = Vector3D::new(1.85, 35.6, 17.8);
    /// let rhs = Vector3D::new(8.5, 42.0, 35.12);
    /// println!("{:?}", Vector3D::dot_product(lhs, rhs));
    /// ```
    fn dot_product(lhs: Vector3D, rhs: Vector3D) -> f64 {
        (lhs.magnitude() * rhs.magnitude()) * (Vector3D::angle(lhs, rhs)).cos()
    }
}

impl Add for Vector3D {
    type Output = Vector3D;

    fn add(self, rhs: Vector3D) -> Vector3D {
        Vector3D::new(self.X + rhs.X, self.Y + rhs.Y, self.Z + rhs.Z)
    }
}

impl Sub for Vector3D {
    type Output = Vector3D;

    fn sub(self, rhs: Vector3D) -> Vector3D {
        Vector3D::new(self.X - rhs.X, self.Y - rhs.Y, self.Z - rhs.Z)
    }
}

impl Mul for Vector3D {
    type Output = Vector3D;

    fn mul(self, rhs: Vector3D) -> Vector3D {
        Vector3D::new(self.X * rhs.X, self.Y * rhs.Y, self.Z * rhs.Z)
    }
}

impl Div for Vector3D {
    type Output = Vector3D;

    fn div(self, rhs: Vector3D) -> Vector3D {
        Vector3D::new(self.X / rhs.X, self.Y / rhs.Y, self.Z / rhs.Z)
    }
}

impl BitOr for Vector3D {
    type Output = f64;

    fn bitor(self, rhs: Vector3D) -> f64 {
        (self.X * rhs.X) + (self.Y * rhs.Y) + (self.Z * rhs.Z)
    }
}
